#! /bin/sh
# ----------------------------------------------------------------------
#  RAPPTURE PROGRAM EXECUTION
#
#  This script implements the -execute option for Rappture.  It
#  provides a way to run a specific Rappture simulation without
#  invoking the Rappture GUI.  Instead, it takes a driver file with
#  the required parameters, submits that for execution, and then
#  returns the run.xml file.  If the -tool file is specified, then
#  it double-checks the driver against the tool to make sure that
#  the requested tool and version are compatible.
#
#  This is normally invoked by launcher.tcl, so it expects $driverxml
#  and $toolxml variable to be already set.
#
#  USAGE: execute.tcl
# ======================================================================
#  AUTHORS:  Michael McLennan, Purdue University
#  Copyright (c) 2004-2012  HUBzero Foundation, LLC
#
#  See the file "license.terms" for information on usage and
#  redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# ======================================================================
#\
exec tclsh "$0" $*
# ----------------------------------------------------------------------
# tclsh executes everything from here on...

# bring in the Rappture object system
package require Rappture
Rappture::objects::init
Rappture::resources::load

# Parse command line options
# ----------------------------------------------------------------------
Rappture::getopts argv params {
    value -tool ""
    value -status ""
    value -stdOutput ""
    value -stdError ""
    value -output ""
    value -cleanup no
    value -xmlSource ""
}

# load the XML info in the driver file
if {[catch {Rappture::library $driverxml} result]} {
    puts stderr "ERROR while loading driver file \"$driverxml\""
    puts stderr $result
    exit 1
}
set driverobj $result

# If tool.xml is not specified, try to find it the way Rappture would.
if {$params(-tool) ne ""} {
    set toolxml $params(-tool)
}
if {$toolxml eq ""} {
    if {[file isfile tool.xml]} {
        set toolxml [file normalize tool.xml]
    } elseif {[file isfile [file join rappture tool.xml]]} {
        set toolxml [file normalize [file join rappture tool.xml]]
    }
}

# If there's still no tool.xml, then see if we can find tooldir in driver
if {$toolxml eq ""} {
    set tooldir [$driverobj get tool.version.application.directory(tool)]
    if {$tooldir eq ""} {
        puts stderr "ERROR: missing -tool option, and driver file doesn't contain sufficient detail to locate the desired tool."
        exit 1
    }

    set toolxml [file join $tooldir tool.xml]
    if {![file exists $toolxml]} {
        puts stderr "ERROR: missing tool.xml file \"$toolxml\""
        exit 1
    }
}

set installdir [file dirname [file normalize $toolxml]]
set toolobj [Rappture::library $toolxml]

# copy structure.current.fields and structure.current.components
foreach path [Rappture::entities -as path $driverobj input] {
    if       {[$driverobj element -as type $path.current] eq ""} {
# if element has no current tag copy it
        $toolobj copy $path from $driverobj $path
    } elseif {[$driverobj get $path.current] eq ""} {
# if element has current tag and it is empty and toolobj is missing the element
        if {[$toolobj element -as type $path.current] eq ""} {
            $toolobj copy $path.current from $driverobj $path.current
        }
    }
}
if {[$driverobj element -as type meta] ne ""} {
    $toolobj copy meta from $driverobj meta
}

if {$params(-xmlSource) != ""} {
    set TaskObj [Rappture::Task ::#auto $toolobj $installdir -xmlSource $params(-xmlSource)]
} else {
    set TaskObj [Rappture::Task ::#auto $toolobj $installdir]
}
set LogStdoutFid ""
set LogStderrFid ""

# Define some things that we need for logging status...
# ----------------------------------------------------------------------
proc log_error {message} {
    global LogStderrFid

    if {$LogStderrFid ne ""} {
        puts -nonewline $LogStderrFid $message
    }
}

proc log_output {message} {
    global LogStdoutFid

    if {$LogStdoutFid ne ""} {
        if {$LogStdoutFid eq "stdout"} {
            puts -nonewline $LogStdoutFid $message
            flush $LogStdoutFid
        } else {
            #
            # Scan through and pick out any =RAPPTURE-PROGRESS=> messages.
            #
            set percent ""
            while {[regexp -indices \
                    {=RAPPTURE-PROGRESS=> *([-+]?[0-9]+) +([^\n]*)(\n|$)} $message \
                     match percent mesg]} {
    
                foreach {i0 i1} $percent break
                set percent [string range $message $i0 $i1]
    
                foreach {i0 i1} $mesg break
                set mesg [string range $message $i0 $i1]
    
                foreach {i0 i1} $match break
                set message [string replace $message $i0 $i1]
            }
            if {$percent ne ""} {
                # report the last percent progress found
                log_append progress "$percent% - $mesg"
            }
        }
    }
}

# Actually write to the log file
proc log_append {level message} {
    global LogStdoutFid

    if {$LogStdoutFid ne ""} {
        set date [clock format [clock seconds] -format {%Y-%m-%dT%H:%M:%S%z}]
        set host [info hostname]
        puts $LogStdoutFid "$date $host rappture [pid] \[$level\] $message"
        flush $LogStdoutFid
    }
}

# Actually write to the log file
proc log_stats {args} {
    set line ""
    foreach {key val} $args {
        append line "$key=$val "
    }
    log_append usage $line
}

# Apply effects of all other command line options
# ----------------------------------------------------------------------
if {$params(-stdOutput) ne ""} {
    if {$params(-stdOutput) eq "@stdout"} {
        set LogStdoutFid "stdout"
    } else {
        set LogStdoutFid [open $params(-stdOutput) w]
    }
    lappend args -stdout log_output
}
if {$params(-stdError) ne ""} {
    if {$params(-stdError) eq "@stderr"} {
        set LogStderrFid "stderr"
    } else {
        set LogStderrFid [open $params(-stdError) w]
    }
    lappend args -stderr log_error
}

if {$params(-status) ne ""} {
    if {$params(-status) ne "@default"} {
        if {$LogStdoutFid eq ""} {
            set LogStdoutFid [open $params(-status) w]
        }
        $TaskObj configure -logger {log_append status} -jobstats log_stats
    }
}

if {$params(-output) eq ""} {
    # no output? then run quietly and don't try to save results
    $TaskObj configure -jobstats "" -resultdir ""
}

# Transfer input values from driver to TaskObj, and then run.
# ----------------------------------------------------------------------

# copy inputs from the test into the run file
$TaskObj reset
foreach path [Rappture::entities -as path $driverobj input] {
    if {[$driverobj element -as type $path.current] ne ""} {
        set currentVal [$driverobj get $path.current]
# do not replace empty paths - in particular structures
        if {$currentVal ne {}} {
            lappend args $path $currentVal
        }
    }
}

if {$params(-status) ne ""} {
    if {$params(-status) ne "@default"} {
        # recording status? then look through output for progress messages
        lappend args -output log_output
    }
}

# run the desired case...
foreach {status result} [eval $TaskObj run $args] break

if {$status == 0 && [Rappture::library isvalid $result]} {
    set runxml $result
    $runxml put output.status ok
} else {
    # build a run file for the result output
    set info "<?xml version=\"1.0\"?>\n[$driverobj xml]"
    set runxml [Rappture::LibraryObj ::#auto $info]
    $runxml put output.log $result
    $runxml put output.status failed
}

# Handle output
# ----------------------------------------------------------------------
switch -- $params(-output) {
    "" {
        # no output file -- write to stdout
        puts "<?xml version=\"1.0\"?>\n[$runxml xml]"
    }
    "@default" {
        # do the usual Rappture thing -- move to results dir
        # but ignore any errors if it fails
        catch {$TaskObj save $runxml}
    }
    default {
        # save to the specified file
        $TaskObj save $runxml $params(-output)
    }
}

if {$params(-cleanup)} {
    file delete -force -- $driverxml
}

if {$params(-status) ne ""} {
    if {$params(-status) ne "@default"} {
       log_append status "exit $status"
    }
}
exit $status
